#include <hikari/cursor.h>

#include <assert.h>
#include <errno.h>
#include <math.h>

#include <wlr/types/wlr_cursor.h>
#include <wlr/types/wlr_pointer_constraints_v1.h>
#include <wlr/types/wlr_seat.h>

#include <hikari/binding.h>
#include <hikari/binding_config.h>
#include <hikari/memory.h>
#include <hikari/output.h>
#include <hikari/server.h>
#include <hikari/view.h>

static void
motion_absolute_handler(struct wl_listener *listener, void *data);

static void
frame_handler(struct wl_listener *listener, void *data);

static void
motion_handler(struct wl_listener *listener, void *data);

static void
button_handler(struct wl_listener *listener, void *data);

static void
axis_handler(struct wl_listener *listener, void *data);

static void
request_set_cursor_handler(struct wl_listener *listener, void *data);

static void
surface_destroy_handler(struct wl_listener *listener, void *data);

static void
check_constraint_region(struct hikari_cursor *cursor);

static void
constraint_commit_handler(struct wl_listener *listener, void *data);

static void
warp_to_constraint_cursor_hint(struct hikari_cursor *cursor);

static unsigned long
get_cursor_size(void)
{
  const char *cursor_size = getenv("XCURSOR_SIZE");

  if (cursor_size != NULL && strlen(cursor_size) > 0) {
    errno = 0;
    char *end;
    unsigned long size = strtoul(cursor_size, &end, 10);
    if (*end == '\0' && errno == 0) {
      return size;
    }
  }

  setenv("XCURSOR_SIZE", "16", 1);

  return 16;
}

static const char *
get_cursor_theme(void)
{
  char *cursor_theme = getenv("XCURSOR_THEME");

  if (cursor_theme != NULL && strlen(cursor_theme) > 0) {
    return cursor_theme;
  }

  setenv("XCURSOR_THEME", "Adwaita", 1);

  return "Adwaita";
}

static void
configure_bindings(struct hikari_cursor *cursor, struct wl_list *bindings)
{
  int nr[256] = { 0 };
  struct hikari_binding_config *binding_config;
  wl_list_for_each (binding_config, bindings, link) {
    nr[binding_config->key.modifiers]++;
  }

  for (int mask = 0; mask < 256; mask++) {
    cursor->bindings[mask].nbindings = nr[mask];
    if (nr[mask] != 0) {
      cursor->bindings[mask].bindings =
          hikari_calloc(nr[mask], sizeof(struct hikari_binding));
    } else {
      cursor->bindings[mask].bindings = NULL;
    }

    nr[mask] = 0;
  }

  wl_list_for_each (binding_config, bindings, link) {
    uint8_t mask = binding_config->key.modifiers;
    struct hikari_binding *binding = &cursor->bindings[mask].bindings[nr[mask]];

    binding->action = &binding_config->action;

    switch (binding_config->key.type) {
      case HIKARI_ACTION_BINDING_KEY_KEYCODE:
        binding->keycode = binding_config->key.value.keycode;
        break;

      case HIKARI_ACTION_BINDING_KEY_KEYSYM:
        assert(false);
        break;
    }

    nr[mask]++;
  }
}

void
hikari_cursor_init(
    struct hikari_cursor *cursor, struct wlr_output_layout *output_layout)
{
  struct wlr_cursor *wlr_cursor = wlr_cursor_create();

  wlr_cursor_attach_output_layout(wlr_cursor, output_layout);

  const char *cursor_theme = get_cursor_theme();
  unsigned long cursor_size = get_cursor_size();

  cursor->cursor_mgr = wlr_xcursor_manager_create(cursor_theme, cursor_size);
  wlr_xcursor_manager_load(cursor->cursor_mgr, 1);

  cursor->wlr_cursor = wlr_cursor;

  wl_list_init(&cursor->surface_destroy.link);
  wl_list_init(&cursor->constraint_commit.link);

  hikari_binding_group_init(cursor->bindings);
}

void
hikari_cursor_configure_bindings(
    struct hikari_cursor *cursor, struct wl_list *bindings)
{
  hikari_binding_group_fini(cursor->bindings);
  hikari_binding_group_init(cursor->bindings);
  configure_bindings(cursor, bindings);
}

void
hikari_cursor_fini(struct hikari_cursor *cursor)
{
  hikari_binding_group_fini(cursor->bindings);
  hikari_cursor_deactivate(cursor);

  wlr_xcursor_manager_destroy(cursor->cursor_mgr);
}

void
hikari_cursor_activate(struct hikari_cursor *cursor)
{
  struct wlr_cursor *wlr_cursor = cursor->wlr_cursor;

  cursor->motion_absolute.notify = motion_absolute_handler;
  wl_signal_add(&wlr_cursor->events.motion_absolute, &cursor->motion_absolute);

  cursor->frame.notify = frame_handler;
  wl_signal_add(&wlr_cursor->events.frame, &cursor->frame);

  cursor->motion.notify = motion_handler;
  wl_signal_add(&wlr_cursor->events.motion, &cursor->motion);

  cursor->button.notify = button_handler;
  wl_signal_add(&wlr_cursor->events.button, &cursor->button);

  cursor->axis.notify = axis_handler;
  wl_signal_add(&wlr_cursor->events.axis, &cursor->axis);

  cursor->request_set_cursor.notify = request_set_cursor_handler;
  wl_signal_add(&hikari_server.seat->events.request_set_cursor,
      &cursor->request_set_cursor);

  hikari_cursor_reset_image(cursor);
}

void
hikari_cursor_deactivate(struct hikari_cursor *cursor)
{
  wl_list_remove(&cursor->motion_absolute.link);
  wl_list_remove(&cursor->frame.link);
  wl_list_remove(&cursor->motion.link);
  wl_list_remove(&cursor->button.link);
  wl_list_remove(&cursor->axis.link);
  wl_list_remove(&cursor->request_set_cursor.link);

  hikari_cursor_set_image(cursor, NULL);
}

void
hikari_cursor_set_image(struct hikari_cursor *cursor, const char *path)
{
  wl_list_remove(&cursor->surface_destroy.link);
  wl_list_init(&cursor->surface_destroy.link);

  if (path != NULL) {
    wlr_cursor_set_xcursor(cursor->wlr_cursor, cursor->cursor_mgr, path);
  } else {
    wlr_cursor_set_xcursor(cursor->wlr_cursor, cursor->cursor_mgr, "default");
  }
}

void
hikari_cursor_center(struct hikari_cursor *cursor,
    struct hikari_output *output,
    struct wlr_box *geometry)
{
  int x = output->geometry.x + geometry->x + geometry->width / 2;
  int y = output->geometry.y + geometry->y + geometry->height / 2;

  hikari_cursor_warp(cursor, x, y);
}

void
hikari_cursor_constrain(
    struct hikari_cursor *cursor, struct wlr_pointer_constraint_v1 *constraint)
{
  if (cursor->active_constraint == constraint) {
    return;
  }

  wl_list_remove(&cursor->constraint_commit.link);
  if (cursor->active_constraint) {
    if (constraint == NULL) {
      warp_to_constraint_cursor_hint(cursor);
    }
    wlr_pointer_constraint_v1_send_deactivated(cursor->active_constraint);
  }

  cursor->active_constraint = constraint;

  if (constraint == NULL) {
    wl_list_init(&cursor->constraint_commit.link);
    return;
  }

  cursor->active_confine_requires_warp = true;

  if (pixman_region32_not_empty(&constraint->current.region)) {
    pixman_region32_intersect(&constraint->region,
        &constraint->surface->input_region,
        &constraint->current.region);
  } else {
    pixman_region32_copy(
        &constraint->region, &constraint->surface->input_region);
  }

  check_constraint_region(cursor);

  wlr_pointer_constraint_v1_send_activated(constraint);

  cursor->constraint_commit.notify = constraint_commit_handler;
  wl_signal_add(
      &constraint->surface->events.commit, &cursor->constraint_commit);
}

static void
motion_absolute_handler(struct wl_listener *listener, void *data)
{
  struct hikari_cursor *cursor =
      wl_container_of(listener, cursor, motion_absolute);

  assert(!hikari_server_in_lock_mode());

  struct wlr_pointer_motion_absolute_event *event = data;

  wlr_cursor_warp_absolute(
      cursor->wlr_cursor, &event->pointer->base, event->x, event->y);

  hikari_server.mode->cursor_move(event->time_msec);
}

static void
frame_handler(struct wl_listener *listener, void *data)
{
  assert(!hikari_server_in_lock_mode());

  wlr_seat_pointer_notify_frame(hikari_server.seat);
}

static void
motion_handler(struct wl_listener *listener, void *data)
{
  struct hikari_cursor *cursor = wl_container_of(listener, cursor, motion);

  assert(!hikari_server_in_lock_mode());

  struct wlr_pointer_motion_event *event = data;

  wlr_cursor_move(cursor->wlr_cursor,
      &event->pointer->base,
      event->delta_x,
      event->delta_y);

  hikari_server.mode->cursor_move(event->time_msec);
}

static void
button_handler(struct wl_listener *listener, void *data)
{
  assert(!hikari_server_in_lock_mode());

  struct hikari_cursor *cursor = wl_container_of(listener, cursor, button);
  struct wlr_pointer_button_event *event = data;

  hikari_server.mode->button_handler(cursor, event);
}

static void
axis_handler(struct wl_listener *listener, void *data)
{
  assert(!hikari_server_in_lock_mode());

  struct wlr_pointer_axis_event *event = data;

  wlr_seat_pointer_notify_axis(hikari_server.seat,
      event->time_msec,
      event->orientation,
      event->delta,
      event->delta_discrete,
      event->source);
}

static void
request_set_cursor_handler(struct wl_listener *listener, void *data)
{
  if (!hikari_server_in_normal_mode()) {
    return;
  }

  struct hikari_cursor *cursor =
      wl_container_of(listener, cursor, request_set_cursor);

  struct hikari_server *server = &hikari_server;
  struct wlr_seat_pointer_request_set_cursor_event *event = data;
  struct wlr_seat *seat = server->seat;

  struct wl_client *focused_client = NULL;
  struct wlr_surface *focused_surface = seat->pointer_state.focused_surface;
  if (focused_surface != NULL) {
    focused_client = wl_resource_get_client(focused_surface->resource);
  }

  if (focused_client == NULL || event->seat_client->client != focused_client) {
    return;
  }

  struct wlr_surface *surface = event->surface;

  wl_list_remove(&cursor->surface_destroy.link);
  if (surface != NULL) {
    cursor->surface_destroy.notify = surface_destroy_handler;
    wl_signal_add(&surface->events.destroy, &cursor->surface_destroy);
  } else {
    wl_list_init(&cursor->surface_destroy.link);
  }

  wlr_cursor_set_surface(
      cursor->wlr_cursor, surface, event->hotspot_x, event->hotspot_y);
}

static void
surface_destroy_handler(struct wl_listener *listener, void *data)
{
  struct hikari_cursor *cursor =
      wl_container_of(listener, cursor, surface_destroy);

  hikari_cursor_reset_image(cursor);
}

static void
check_constraint_region(struct hikari_cursor *cursor)
{
  struct wlr_pointer_constraint_v1 *constraint = cursor->active_constraint;
  pixman_region32_t *region = &constraint->region;
  struct hikari_view *view = view_from_wlr_surface(constraint->surface);
  if (cursor->active_confine_requires_warp && view) {
    cursor->active_confine_requires_warp = false;

    double sx = cursor->wlr_cursor->x - view->current_geometry->x;
    double sy = cursor->wlr_cursor->y - view->current_geometry->y;

    if (!pixman_region32_contains_point(region, floor(sx), floor(sy), NULL)) {
      int nboxes;
      pixman_box32_t *boxes = pixman_region32_rectangles(region, &nboxes);
      if (nboxes > 0) {
        double sx = (boxes[0].x1 + boxes[0].x2) / 2.;
        double sy = (boxes[0].y1 + boxes[0].y2) / 2.;

        wlr_cursor_warp_closest(cursor->wlr_cursor,
            NULL,
            sx + view->current_geometry->x,
            sy + view->current_geometry->y);

        hikari_server_cursor_focus();
      }
    }
  }

  if (constraint->type == WLR_POINTER_CONSTRAINT_V1_CONFINED) {
    pixman_region32_copy(&cursor->confine, region);
  } else {
    pixman_region32_clear(&cursor->confine);
  }
}

static void
constraint_commit_handler(struct wl_listener *listener, void *data)
{
  struct hikari_cursor *cursor =
      wl_container_of(listener, cursor, constraint_commit);
  struct wlr_pointer_constraint_v1 *constraint = cursor->active_constraint;
  assert(constraint->surface == data);

  check_constraint_region(cursor);
}

static void
pointer_constraint_set_region_handler(struct wl_listener *listener, void *data)
{
  struct hikari_pointer_constraint *hikari_constraint =
      wl_container_of(listener, hikari_constraint, set_region);
  struct hikari_cursor *cursor = hikari_constraint->cursor;

  cursor->active_confine_requires_warp = true;
}

static void
warp_to_constraint_cursor_hint(struct hikari_cursor *cursor)
{
  struct wlr_pointer_constraint_v1 *constraint = cursor->active_constraint;

  if (constraint->current.committed &
      WLR_POINTER_CONSTRAINT_V1_STATE_CURSOR_HINT) {
    double sx = constraint->current.cursor_hint.x;
    double sy = constraint->current.cursor_hint.y;

    struct hikari_view *view = view_from_wlr_surface(constraint->surface);

    if (view != NULL) {
      double lx = sx + view->current_geometry->x;
      double ly = sy + view->current_geometry->y;

      wlr_cursor_warp(cursor->wlr_cursor, NULL, lx, ly);
      wlr_seat_pointer_warp(constraint->seat, sx, sy);
    }
  }
}

static void
constraint_destroy_handler(struct wl_listener *listener, void *data)
{
  struct hikari_pointer_constraint *hikari_constraint =
      wl_container_of(listener, hikari_constraint, destroy);
  struct wlr_pointer_constraint_v1 *constraint = data;
  struct hikari_cursor *cursor = hikari_constraint->cursor;

  wl_list_remove(&hikari_constraint->set_region.link);
  wl_list_remove(&hikari_constraint->destroy.link);

  if (cursor->active_constraint == constraint) {
    warp_to_constraint_cursor_hint(cursor);

    if (cursor->constraint_commit.link.next != NULL) {
      wl_list_remove(&cursor->constraint_commit.link);
    }
    wl_list_init(&cursor->constraint_commit.link);
    cursor->active_constraint = NULL;
  }

  free(hikari_constraint);
}

void
pointer_constraint_handler(struct wl_listener *listener, void *data)
{
  struct wlr_pointer_constraint_v1 *constraint = data;

  struct hikari_pointer_constraint *hikari_constraint =
      hikari_malloc(sizeof(struct hikari_pointer_constraint));
  hikari_constraint->cursor = &hikari_server.cursor;
  hikari_constraint->constraint = constraint;

  hikari_constraint->set_region.notify = pointer_constraint_set_region_handler;
  wl_signal_add(&constraint->events.set_region, &hikari_constraint->set_region);

  hikari_constraint->destroy.notify = constraint_destroy_handler;
  wl_signal_add(&constraint->events.destroy, &hikari_constraint->destroy);

  struct hikari_view *focus_view = hikari_server.workspace->focus_view;
  if (focus_view != NULL && focus_view->surface != NULL) {
    if (focus_view->surface == constraint->surface) {
      hikari_cursor_constrain(&hikari_server.cursor, constraint);
    }
  }
}
