#include <hikari/output.h>

#include <libdrm/drm_fourcc.h>

#include <wlr/backend.h>
#include <wlr/types/wlr_output.h>
#include <wlr/util/log.h>
#ifdef HAVE_GAMMACONTROL
#include <wlr/types/wlr_gamma_control_v1.h>
#endif

#include <hikari/memory.h>
#include <hikari/renderer.h>
#include <hikari/server.h>
#ifdef HAVE_XWAYLAND
#include <hikari/view.h>
#endif

static void
hikari_output_set_mode(struct wlr_output *wlr_output,
    struct wlr_output_state *pending,
    struct hikari_mode_config *imode);

static inline void
render_image_to_surface(cairo_surface_t *output,
    cairo_surface_t *image,
    enum hikari_background_fit fit)
{
  cairo_t *cairo = cairo_create(output);
  if (cairo_surface_status(image) != CAIRO_STATUS_SUCCESS) {
    wlr_log(WLR_DEBUG,
        "hikari: in render_image_to_surface(),"
        "cairo_surface_status() failed.");
    return;
  }

  double output_width = cairo_image_surface_get_width(output);
  double output_height = cairo_image_surface_get_height(output);
  double width = cairo_image_surface_get_width(image);
  double height = cairo_image_surface_get_height(image);

  cairo_rectangle(cairo, 0, 0, output_width, output_height);
  cairo_fill(cairo);

  if (fit == HIKARI_BACKGROUND_STRETCH) {
    cairo_scale(cairo, output_width / width, output_height / height);
    cairo_set_source_surface(cairo, image, 0, 0);
  } else if (fit == HIKARI_BACKGROUND_CENTER) {
    cairo_set_source_surface(cairo,
        image,
        output_width / 2 - width / 2,
        output_height / 2 - height / 2);
  } else if (fit == HIKARI_BACKGROUND_TILE) {
    cairo_pattern_t *pattern = cairo_pattern_create_for_surface(image);
    cairo_pattern_set_extend(pattern, CAIRO_EXTEND_REPEAT);
    cairo_set_source(cairo, pattern);
    cairo_pattern_destroy(pattern);
  }

  cairo_paint(cairo);
  cairo_destroy(cairo);
}

void
hikari_output_load_background(struct hikari_output *output,
    const char *path,
    enum hikari_background_fit background_fit)
{
  if (output->background != NULL) {
    wlr_texture_destroy(output->background);
    output->background = NULL;
  }

  assert(output->background == NULL);

  if (path == NULL) {
    goto done;
  }

  cairo_surface_t *image = cairo_image_surface_create_from_png(path);
  if (cairo_surface_status(image) != CAIRO_STATUS_SUCCESS) {
    goto done;
  }

  int output_width = output->geometry.width;
  int output_height = output->geometry.height;
  wlr_log(WLR_DEBUG, "hikari: output->geometry.width: %d", output_width);
  wlr_log(WLR_DEBUG, "hikari: output->geometry.height: %d", output_height);

  cairo_surface_t *output_surface = cairo_image_surface_create(
      CAIRO_FORMAT_ARGB32, output_width, output_height);
  if (cairo_surface_status(image) != CAIRO_STATUS_SUCCESS) {
    cairo_surface_destroy(image);
    goto done;
  }

  render_image_to_surface(output_surface, image, background_fit);

  unsigned char *data = cairo_image_surface_get_data(output_surface);
  int stride = cairo_format_stride_for_width(CAIRO_FORMAT_ARGB32, output_width);

  struct wlr_renderer *renderer = output->wlr_output->renderer;

  output->background = wlr_texture_from_pixels(
      renderer, DRM_FORMAT_ARGB8888, stride, output_width, output_height, data);

  cairo_surface_destroy(image);
  cairo_surface_destroy(output_surface);

done:
  if (output->enabled) {
    hikari_output_damage_whole(output);
  }
}

void
hikari_output_damage_whole(struct hikari_output *output)
{
  assert(output != NULL);

  wlr_damage_ring_add_whole(output->damage_ring);
  wlr_output_schedule_frame(output->wlr_output);
}

static void
hikari_output_needs_frame_handler(struct wl_listener *listener, void *data)
{
  struct hikari_output *output = wl_container_of(listener, output, needs_frame);
  wlr_output_schedule_frame(output->wlr_output);
}

static void
hikari_output_damage_handler(struct wl_listener *listener, void *data)
{
  struct hikari_output *output = wl_container_of(listener, output, damage);
  struct wlr_output_event_damage *event = data;
  if (wlr_damage_ring_add(output->damage_ring, event->damage)) {
    wlr_output_schedule_frame(output->wlr_output);
  }
}

static void
queue_output_config(struct hikari_output_config *output_config,
    struct hikari_output *output,
    struct wlr_output_state *pending)
{
  if (!output->enabled) {
    return;
  }

  struct wlr_output *wlr_output = output->wlr_output;

  wlr_output_state_set_enabled(pending, true);

  if (hikari_output_config_has_mode(output_config)) {
    struct hikari_mode_config mode =
        hikari_output_config_get_mode(output_config);
    hikari_output_set_mode(wlr_output, pending, &mode);
  } else {
    struct wlr_output_mode *mode = wlr_output_preferred_mode(wlr_output);
    if (mode != NULL) {
      wlr_output_state_set_mode(pending, mode);
    }
  }

  if (hikari_output_config_has_adaptive_sync(output_config)) {
    wlr_output_state_set_adaptive_sync_enabled(
        pending, hikari_output_config_get_adaptive_sync(output_config));
  }
}

static bool
apply_output_config(
    struct hikari_output_config *output_config, struct hikari_output *output)
{
  bool result;
  struct wlr_output *wlr_output = output->wlr_output;
  struct wlr_output_state pending;
  wlr_output_state_init(&pending);
  queue_output_config(output_config, output, &pending);
  result = wlr_output_commit_state(wlr_output, &pending);
  wlr_output_state_finish(&pending);
  if (!result) {
    return false;
  }
  return true;
}

static bool
test_output_config(
    struct hikari_output_config *output_config, struct hikari_output *output)
{
  bool result;
  struct wlr_output_state pending;
  wlr_output_state_init(&pending);
  queue_output_config(output_config, output, &pending);
  result = wlr_output_test_state(output->wlr_output, &pending);
  wlr_output_state_finish(&pending);
  return result;
}

void
hikari_output_disable(struct hikari_output *output)
{
  wlr_log(WLR_DEBUG,
      "hikari: %s called for %s.",
      __func__,
      output->wlr_output->name);
  assert(output != NULL);

  if (!output->enabled) {
    return;
  }

  bool outputs_enabled = false;
  struct hikari_output *each_output;
  struct wlr_output *wlr_output = output->wlr_output;

  wlr_seat_keyboard_notify_clear_focus(hikari_server.seat);
  wlr_seat_pointer_notify_clear_focus(hikari_server.seat);

  wl_list_remove(&output->damage.link);
  wl_list_init(&output->damage.link);
  wl_list_remove(&output->frame.link);
  wl_list_init(&output->frame.link);
  wl_list_remove(&output->needs_frame.link);
  wl_list_init(&output->needs_frame.link);

  struct wlr_output_state output_state;
  wlr_output_state_init(&output_state);
  wlr_output_state_set_enabled(&output_state, false);
  wlr_output_commit_state(wlr_output, &output_state);
  wlr_output_state_finish(&output_state);

  output->enabled = false;
  wl_list_for_each (each_output, &hikari_server.outputs, server_outputs) {
    outputs_enabled |= each_output->enabled;
  }
  if (!outputs_enabled) {
    hikari_output_power_state_enter();
    hikari_server.output_power_state.outputs_disabled = true;
  }
}

void
hikari_output_enable(struct hikari_output *output)
{
  wlr_log(WLR_DEBUG,
      "hikari: %s called for %s.",
      __func__,
      output->wlr_output->name);
  assert(output != NULL);

  if (output->enabled) {
    return;
  }

  struct wlr_output *wlr_output = output->wlr_output;

  wlr_seat_keyboard_notify_clear_focus(hikari_server.seat);
  wlr_seat_pointer_notify_clear_focus(hikari_server.seat);

  wl_list_remove(&output->frame.link);
  output->frame.notify = hikari_renderer_frame_handler;
  wl_signal_add(&wlr_output->events.frame, &output->frame);
  output->needs_frame.notify = hikari_output_needs_frame_handler;
  wl_signal_add(&wlr_output->events.needs_frame, &output->needs_frame);
  output->damage.notify = hikari_output_damage_handler;
  wl_signal_add(&wlr_output->events.damage, &output->damage);

  struct hikari_output_config *output_config =
      hikari_configuration_resolve_output_config(
          hikari_configuration, wlr_output->name);

  struct wlr_output_state output_state;
  wlr_output_state_init(&output_state);
  wlr_output_state_set_enabled(&output_state, true);
  if (!hikari_output_config_has_mode(output_config)) {
    struct wlr_output_mode *mode = wlr_output_preferred_mode(wlr_output);
    if (mode != NULL) {
      wlr_output_state_set_mode(&output_state, mode);
    }
  }
  wlr_output_commit_state(wlr_output, &output_state);
  wlr_output_state_finish(&output_state);
  output->enabled = true;

  apply_output_config(output_config, output);
  hikari_cursor_reset_image(&hikari_server.cursor);
  hikari_output_damage_whole(output);

  if (hikari_server.output_power_state.outputs_disabled) {
    hikari_output_power_state_fini();
    hikari_server.output_power_state.outputs_disabled = false;
  }
}

static void
output_geometry(struct hikari_output *output)
{
  struct wlr_box output_box;
  wlr_output_layout_get_box(
      hikari_server.output_layout, output->wlr_output, &output_box);

  output->geometry.x = output_box.x;
  output->geometry.y = output_box.y;
  output->geometry.width = output_box.width;
  output->geometry.height = output_box.height;

  output->usable_area = (struct wlr_box){
    .x = 0, .y = 0, .width = output_box.width, .height = output_box.height
  };
}

#ifdef HAVE_LAYERSHELL
static void
close_layers(struct wl_list *layers)
{
  struct hikari_layer *layer, *layer_temp;
  wl_list_for_each_safe (layer, layer_temp, layers, layer_surfaces) {
    wlr_layer_surface_v1_destroy(layer->surface);
    layer->output = NULL;
  }
}
#endif

static void
destroy_handler(struct wl_listener *listener, void *data)
{
  struct hikari_output *output = wl_container_of(listener, output, destroy);

  wlr_log(WLR_DEBUG, "hikari: destroy output %p", output);

  hikari_output_fini(output);
  hikari_free(output);
}

static void
update_output_manager_config(struct hikari_server *server)
{
  struct wlr_output_configuration_v1 *config =
      wlr_output_configuration_v1_create();

  struct hikari_output *output;
  wl_list_for_each (output, &server->outputs, server_outputs) {
    struct wlr_output_configuration_head_v1 *config_head =
        wlr_output_configuration_head_v1_create(config, output->wlr_output);
  }
  wlr_output_manager_v1_set_configuration(server->output_manager, config);
}

#ifdef HAVE_GAMMACONTROL
static void
set_gamma_handler(struct wl_listener *listener, void *data)
{
  wlr_log(WLR_DEBUG, "hikari: set_gamma_handler().");

  struct hikari_output *output = wl_container_of(listener, output, set_gamma);
  struct wlr_gamma_control_manager_v1_set_gamma_event *event = data;
  struct wlr_output_state state;
  wlr_output_state_init(&state);

  if (!wlr_gamma_control_v1_apply(event->control, &state)) {
    wlr_gamma_control_v1_send_failed_and_destroy(event->control);
    wlr_output_state_finish(&state);
    return;
  }

  wlr_output_commit_state(event->output, &state);
  hikari_output_schedule_frame(output);
}
#endif

void
hikari_output_init(struct hikari_output *output, struct wlr_output *wlr_output)
{
  assert(!output->enabled);

  bool noop = wlr_output->backend == hikari_server.noop_backend;

  output->wlr_output = wlr_output;
  struct wlr_damage_ring *damage_ring =
      hikari_calloc(1, sizeof(struct wlr_damage_ring));
  wlr_damage_ring_init(damage_ring);
  output->damage_ring = damage_ring;
  output->background = NULL;
  output->enabled = false;
  output->workspace = hikari_malloc(sizeof(struct hikari_workspace));

#ifdef HAVE_XWAYLAND
  wl_list_init(&output->unmanaged_xwayland_views);
#endif
  wl_list_init(&output->views);

#ifdef HAVE_LAYERSHELL
  wl_list_init(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY]);
  wl_list_init(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_TOP]);
  wl_list_init(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM]);
  wl_list_init(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND]);
#endif

  hikari_workspace_init(output->workspace, output);

  wlr_output->data = output;

  output->destroy.notify = destroy_handler;
  wl_signal_add(&wlr_output->events.destroy, &output->destroy);

  if (!noop) {
    bool first = wl_list_empty(&hikari_server.outputs);

    wl_list_insert(&hikari_server.outputs, &output->server_outputs);

#ifdef HAVE_GAMMACONTROL
    struct hikari_server *server = &hikari_server;
    output->set_gamma.notify = set_gamma_handler;
    wl_signal_add(
        &server->gamma_control_manager->events.set_gamma, &output->set_gamma);
#endif

    wl_list_init(&output->damage.link);
    wl_list_init(&output->frame.link);
    wl_list_init(&output->needs_frame.link);

    if (!hikari_server_in_lock_mode()) {
      hikari_output_enable(output);
    } else if (hikari_lock_mode_are_outputs_disabled(
                   &hikari_server.lock_mode)) {
      hikari_output_disable(output);
    }

    struct hikari_output_config *output_config =
        hikari_configuration_resolve_output_config(
            hikari_configuration, wlr_output->name);

    if (output_config != NULL && output_config->position.value.type ==
                                     HIKARI_POSITION_CONFIG_TYPE_ABSOLUTE) {
      int x = output_config->position.value.config.absolute.x;
      int y = output_config->position.value.config.absolute.y;

      wlr_output_layout_add(hikari_server.output_layout, wlr_output, x, y);
    } else {
      wlr_output_layout_add_auto(hikari_server.output_layout, wlr_output);
    }

    output_geometry(output);
    update_output_manager_config(server);

    if (first) {
      hikari_workspace_merge(
          hikari_server.noop_output->workspace, output->workspace);
      hikari_workspace_focus_view(output->workspace, NULL);
    }
  }
}

void
hikari_output_fini(struct hikari_output *output)
{
  bool noop = output->wlr_output->backend == hikari_server.noop_backend;

#ifdef HAVE_LAYERSHELL
  close_layers(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY]);
  close_layers(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_TOP]);
  close_layers(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM]);
  close_layers(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND]);
#endif

  hikari_output_disable(output);

  wl_list_remove(&output->destroy.link);

  struct hikari_workspace *workspace = output->workspace;

  if (!noop) {
    struct hikari_workspace *merge_workspace;
    struct hikari_workspace *next_workspace = hikari_workspace_next(workspace);

    wlr_texture_destroy(output->background);

    if (workspace != next_workspace) {
      merge_workspace = next_workspace;
    } else {
      merge_workspace = hikari_server.noop_output->workspace;
    }

    hikari_workspace_merge(workspace, merge_workspace);

    if (!hikari_server_in_lock_mode()) {
      if (!hikari_server_in_normal_mode()) {
        hikari_server_enter_normal_mode(NULL);
      }

      hikari_workspace_focus_view(merge_workspace, NULL);
    } else {
      merge_workspace->focus_view = NULL;
      hikari_server.workspace = merge_workspace;
    }

    wl_list_remove(&output->server_outputs);

    update_output_manager_config(&hikari_server);
  } else {
    hikari_server.workspace = NULL;
  }

  wlr_damage_ring_finish(output->damage_ring);
  hikari_free(output->damage_ring);
  output->damage_ring = NULL;
  hikari_workspace_fini(workspace);
  hikari_free(workspace);
}

void
hikari_output_move(struct hikari_output *output, double lx, double ly)
{
  wlr_output_layout_add(
      hikari_server.output_layout, output->wlr_output, lx, ly);
  hikari_output_schedule_frame(output); // XXX needless?
}

#define CYCLE_OUTPUT(name)                                                     \
  struct hikari_output *hikari_output_##name(struct hikari_output *output)     \
  {                                                                            \
    if (wl_list_empty(&hikari_server.outputs)) {                               \
      return NULL;                                                             \
    }                                                                          \
                                                                               \
    struct wl_list *name = output->server_outputs.name;                        \
                                                                               \
    if (name == &hikari_server.outputs) {                                      \
      name = hikari_server.outputs.name;                                       \
    }                                                                          \
                                                                               \
    struct hikari_output *name##_output =                                      \
        wl_container_of(name, name##_output, server_outputs);                  \
                                                                               \
    return name##_output;                                                      \
  }

CYCLE_OUTPUT(next)
CYCLE_OUTPUT(prev)
#undef CYCLE_OUTPUT

#ifdef HAVE_XWAYLAND
void
hikari_output_rearrange_xwayland_views(struct hikari_output *output)
{
  struct hikari_view *view;
  wl_list_for_each (view, &output->views, output_views) {
    if (view->move != NULL) {
      struct wlr_box *geometry = hikari_view_geometry(view);
      view->move(view, geometry->x, geometry->y);
      wlr_damage_ring_add_box(output->damage_ring, geometry);
      hikari_output_schedule_frame(output);
    }
  }
}
#endif

static void
hikari_output_set_mode(struct wlr_output *wlr_output,
    struct wlr_output_state *pending,
    struct hikari_mode_config *imode)
{
  int refresh_mhz = roundf(imode->refresh_rate * 1000);
  struct wlr_output_mode *mode, *best = NULL;
  int best_diff_mhz = INT_MAX;

  if (wl_list_empty(&wlr_output->modes) || imode->custom) {
    wlr_log(WLR_INFO,
        "hikari: assigning custom mode (%dx%d@%.3fHz) to %s",
        imode->width,
        imode->height,
        imode->refresh_rate,
        wlr_output->name);
    if (refresh_mhz <= 0) {
      refresh_mhz = INT_MAX;
    }
    wlr_output_state_set_custom_mode(
        pending, imode->width, imode->height, refresh_mhz);
    return;
  }

  wl_list_for_each (mode, &wlr_output->modes, link) {
    if (mode->width == imode->width && mode->height == imode->height) {
      if (imode->refresh_rate == 0.0) {
        best = mode;
        break;
      }
      int diff_mhz = abs(mode->refresh - refresh_mhz);
      if (diff_mhz < best_diff_mhz) {
        best_diff_mhz = diff_mhz;
        best = mode;
        if (best_diff_mhz == 0) {
          break;
        }
      }
    }
  }
  if (best) {
    wlr_log(WLR_INFO,
        "hikari: assigning configured mode (%dx%d@%.3fHz) to %s",
        best->width,
        best->height,
        best->refresh / 1000.f,
        wlr_output->name);
  } else {
    best = wlr_output_preferred_mode(wlr_output);
    wlr_log(WLR_INFO,
        "hikari: configured mode (%dx%d@%.3fHz) not available, "
        "applying preferred mode (%dx%d@%.3fHz) to %s",
        imode->width,
        imode->height,
        imode->refresh_rate,
        best->width,
        best->height,
        best->refresh / 1000.f,
        wlr_output->name);
  }
  wlr_output_state_set_mode(pending, best);
}

void
output_manager_apply_handler(struct wl_listener *listener, void *data)
{
  struct hikari_server *server =
      wl_container_of(listener, server, output_manager_apply);
  struct wlr_output_configuration_v1 *config = data;

  struct wlr_output_configuration_head_v1 *config_head;
  bool result = true;
  wl_list_for_each (config_head, &config->heads, link) {
    struct wlr_output *wlr_output = config_head->state.output;
    struct hikari_output *output = wlr_output->data;

    struct hikari_output_config *output_config =
        hikari_configuration_resolve_output_config(
            hikari_configuration, wlr_output->name);

    if (config_head->state.enabled) {
      hikari_output_enable(output);
    } else {
      hikari_output_disable(output);
    }

    struct hikari_mode_config mode, current_mode;
    struct wlr_output_mode *wlr_mode;
    if (config_head->state.mode != NULL) {
      wlr_mode = config_head->state.mode;
      mode.width = wlr_mode->width;
      mode.height = wlr_mode->height;
      mode.refresh_rate = wlr_mode->refresh / 1000.f;
      mode.custom = false;
    } else if (config_head->state.custom_mode.width > 0 &&
               config_head->state.custom_mode.height > 0) {
      mode.width = config_head->state.custom_mode.width;
      mode.height = config_head->state.custom_mode.height;
      mode.refresh_rate = config_head->state.custom_mode.refresh / 1000.f;
      mode.custom = true;
    } else if (!hikari_output_config_has_mode(output_config)) {
      wlr_mode = wlr_output_preferred_mode(wlr_output);
      mode.width = wlr_mode->width;
      mode.height = wlr_mode->height;
      mode.refresh_rate = wlr_mode->refresh / 1000.f;
      mode.custom = false;
    }
    if (hikari_output_config_has_mode(output_config)) {
      current_mode = hikari_output_config_get_mode(output_config);
    }
    if ((mode.width > 0 && mode.height > 0) &&
        (mode.width != current_mode.width ||
            mode.height != current_mode.height ||
            mode.refresh_rate != current_mode.refresh_rate)) {
      hikari_output_config_set_mode(output_config, mode);
    }

    bool current_adaptive_sync = false;
    if (hikari_output_config_has_adaptive_sync(output_config)) {
      current_adaptive_sync =
          hikari_output_config_get_adaptive_sync(output_config);
    }
    if (current_adaptive_sync != config_head->state.adaptive_sync_enabled) {
      hikari_output_config_set_adaptive_sync(
          output_config, config_head->state.adaptive_sync_enabled);
    }

    result &= apply_output_config(output_config, output);
  }

  if (result) {
    wlr_output_configuration_v1_send_succeeded(config);
  } else {
    wlr_output_configuration_v1_send_failed(config);
  }
  wlr_output_configuration_v1_destroy(config);
  update_output_manager_config(server);
}

void
output_manager_test_handler(struct wl_listener *listener, void *data)
{
  struct hikari_server *server =
      wl_container_of(listener, server, output_manager_test);
  struct wlr_output_configuration_v1 *config = data;

  struct wlr_output_configuration_head_v1 *config_head;
  bool result = true;
  struct hikari_output_config *output_config =
      hikari_malloc(sizeof(struct hikari_output_config));
  wl_list_for_each (config_head, &config->heads, link) {
    struct wlr_output *wlr_output = config_head->state.output;
    struct hikari_output *output = wlr_output->data;

    hikari_output_config_init(output_config, wlr_output->name);

    struct hikari_mode_config mode;
    struct wlr_output_mode *wlr_mode;
    if (config_head->state.mode != NULL) {
      wlr_mode = config_head->state.mode;
      mode.width = wlr_mode->width;
      mode.height = wlr_mode->height;
      mode.refresh_rate = wlr_mode->refresh / 1000.f;
      mode.custom = false;
    } else {
      mode.width = config_head->state.custom_mode.width;
      mode.height = config_head->state.custom_mode.height;
      mode.refresh_rate = config_head->state.custom_mode.refresh / 1000.f;
      mode.custom = true;
    }
    hikari_output_config_set_mode(output_config, mode);

    hikari_output_config_set_adaptive_sync(
        output_config, config_head->state.adaptive_sync_enabled);

    result &= test_output_config(output_config, output);
  }

  if (result) {
    wlr_output_configuration_v1_send_succeeded(config);
  } else {
    wlr_output_configuration_v1_send_failed(config);
  }
  wlr_output_configuration_v1_destroy(config);
  hikari_free(output_config);
}
